/*
  This file is part of KDDockWidgets.

  SPDX-FileCopyrightText: 2019 Klarälvdalens Datakonsult AB, a KDAB Group company <info@kdab.com>
  Author: Sérgio Martins <sergio.martins@kdab.com>

  SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only

  Contact KDAB at <info@kdab.com> for commercial licensing options.
*/

/**
 * @file
 * @brief Namespace-level methods related to registering QML types.
 *
 * @author Sérgio Martins \<sergio.martins@kdab.com\>
 */

#ifndef KD_QMLTYPES_H
#define KD_QMLTYPES_H

namespace KDDockWidgets {
// This is only called for:
// Qt5
// Qt6 if KDDockWidgets_QML_MODULE=OFF
void registerQmlTypes();
}

#endif
