/**
  This file was auto-generated by the Flutter SDK

  SPDX-FileCopyrightText: 2014 The Flutter Authors.
  SPDX-License-Identifier: BSD-3-Clause
*/

import FlutterMacOS
import Cocoa
import XCTest

class RunnerTests: XCTestCase {

  func testExample() {
    // If you add code to the Runner application, consider adding tests here.
    // See https://developer.apple.com/documentation/xctest for more information about using XCTest.
  }

}
