/**
  This file was auto-generated by the Flutter SDK

  SPDX-FileCopyrightText: 2014 The Flutter Authors.
  SPDX-License-Identifier: BSD-3-Clause
*/

import Cocoa
import FlutterMacOS

@main
class AppDelegate: FlutterAppDelegate {
  override func applicationShouldTerminateAfterLastWindowClosed(_ sender: NSApplication) -> Bool {
    return true
  }
}
